/*
 * Copyright (C) 2016 Socionext Inc.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <libfdt.h>
#include <linux/io.h>

#include "../soc-info.h"
#include "local-init.h"

DECLARE_GLOBAL_DATA_PTR;

/*
 * Unneeded for the latest kernel.
 * Ugly workaround here to support Linux 4.4 locally used in Socionext.
 */
static void uniphier_setup_xirq(void)
{
	const void *fdt = gd->fdt_blob;
	int soc_node, aidet_node;
	const fdt32_t *val;
	unsigned long aidet_base;
	u32 tmp;

	soc_node = fdt_path_offset(fdt, "/soc");
	if (soc_node < 0)
		return;

	aidet_node = fdt_subnode_offset_namelen(fdt, soc_node, "aidet", 5);
	if (aidet_node < 0)
		return;

	val = fdt_getprop(fdt, aidet_node, "reg", NULL);
	if (!val)
		return;

	aidet_base = fdt32_to_cpu(*val);

	tmp = readl(aidet_base + 8);	/* AIDET DETCONFR2 */
	tmp |= 0x00ff0000;		/* Set XIRQ0-7 low active */
	writel(tmp, aidet_base + 8);

	tmp = readl(0x55000090);	/* IRQCTL */
	tmp |= 0x000000ff;
	writel(tmp, 0x55000090);
}

int board_early_init_r(void)
{
	switch (uniphier_get_soc_id()) {
#if defined(CONFIG_ARCH_UNIPHIER_LD11)
	case UNIPHIER_LD11_ID:
		uniphier_ld11_local_init();	/* LD11 specific */
		uniphier_ld11_umc_funcsel();
		uniphier_ld11_ave_drvinit();
		break;
#endif
#if defined(CONFIG_ARCH_UNIPHIER_LD20)
	case UNIPHIER_LD20_ID:
		uniphier_ld20_local_init();	/* LD20 specific */
		uniphier_ld20_umc_funcsel();
		uniphier_ld20_ave_drvinit();
		break;
#endif
#if defined(CONFIG_ARCH_UNIPHIER_PXS3)
	case UNIPHIER_PXS3_ID:
		uniphier_pxs3_local_init();	/* PXs3 specific */
		uniphier_pxs3_umc_funcsel();
		break;
#endif
	default:
		break;
	}

	uniphier_setup_xirq();

	return 0;
}
